# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TaskManager handles all events and information related to the task."""

import ast
import copy
import datetime
import json
import queue
import re
import threading
from typing import Any, Optional, Callable, Union, Iterable, Tuple
from typing import Dict, List, Set, Pattern
import functools
import itertools
import operator

from absl import logging
from android_env.components import adb_controller as adb_control

from android_env.components import app_screen_checker
from android_env.components import dumpsys_thread
from android_env.components import errors

from android_env.components import log_stream as log_stream_lib
from android_env.components import logcat_thread

# zdy
from android_env.components.tools import naive_functions
from android_env.components.tools.types import TextModel, IconModel
#from android_env.proto import emulator_controller_pb2
#from android_env.proto import emulator_controller_pb2_grpc
from android_env.components import screen_analyzer_thread

from android_env.components import setup_step_interpreter
from android_env.proto import task_pb2
from android_env.components import event_listeners # zdy
import numpy as np
import lxml.etree

class TaskManager():
  """Handles all events and information related to the task."""

  non_pritable_char_mapping = { "\b": "KEYCODE_DEL"
                              , "\x1b": "KEYCODE_ESCAPE"
                              , "\n": "KEYCODE_ENTER"
                              , "\r": "KEYCODE_ENTER"
                              , "\t": "KEYCODE_TAB"
                              , "`": "KEYCODE_GRAVE"
                              }

  def __init__( self
              , task: task_pb2.Task
              , start_token_mark: str = ""
              , non_start_token_mark: str = "##"
              , special_token_pattern: str = r"\[\w+\]"
              , fix_vocabulary_to: Optional[Iterable[str]] = None
              , max_bad_states: int = 3
              , dumpsys_check_frequency: int = 150
              , max_failed_current_activity: int = 10
              , text_model: TextModel = naive_functions
              , icon_model: IconModel = naive_functions
              ):
    #  method `__init__` {{{ # 
    """Controls task-relevant events and information.

    Args:
      task: A task proto defining the RL task.

      start_token_mark: str as the mark for the start subwords such as "Ġ"
        (\\u0120) for GPT subword
      non_start_token_mark: str as the mark for the non-start subwords such as
        "##" for BERT subword
      special_token_pattern: str as the pattern of the special non-printable
        tokens such as "[CLS]" for BERT tokens
      fix_vocabulary_to: iterable of str or none as the another assigned
        vocabulary rather than the small one in the `task`

      max_bad_states: How many bad states in a row are allowed before a restart
        of the simulator is triggered.
      dumpsys_check_frequency: Frequency, in steps, at which to check
        current_activity and view hierarchy
      max_failed_current_activity: The maximum number of tries for extracting
        the current activity before forcing the episode to restart.

      text_model: TextModel
      icon_model: IconModel
    """
    self._task: task_pb2.Task = task

    self._start_token_mark: str = start_token_mark
    self._non_start_token_mark: str = non_start_token_mark
    self._non_start_mark_first: bool = True\
                                     if self._non_start_token_mark.startswith(self._start_token_mark)\
                                     else False
    self._special_token_pattern: Pattern[str] = re.compile(special_token_pattern)

    self._max_bad_states: int = max_bad_states
    self._dumpsys_check_frequency: int = dumpsys_check_frequency
    self._max_failed_current_activity: int = max_failed_current_activity

    self._lock: threading.Lock = threading.Lock()
    self._extras_max_buffer_size: int = 100
    self._adb_controller: Optional[adb_control.AdbController] = None
    #self._emulator_stub: Optional[emula = None
    #self._image_format = None
    self._setup_step_interpreter:\
        Optional[setup_step_interpreter.SetupStepInterpreter] = None

    self._vocabulary: List[str] = list(fix_vocabulary_to) if fix_vocabulary_to is not None\
                                                          else list(self._task.vocabulary)

    # Logging settings
    self._log_dict: Dict[str, int] = {
        'reset_count_step_timeout': 0,
        'reset_count_player_exited': 0,
        'reset_count_episode_end': 0,
        'reset_count_max_duration_reached': 0,
        'restart_count_max_bad_states': 0,
    }

    # zdy
    #  Event Infrastructures {{{ # 
    self._events_with_id: Dict[int, event_listeners.Event] = {}
    self._events_in_source: Dict[int, List[event_listeners.Event]] = {}
    self._events_in_need: Dict[int, List[event_listeners.Event]] = {}

    # Event Sources
    self._text_events: List[event_listeners.TextEvent] = []
    self._icon_events: List[event_listeners.IconRecogEvent] = []
    self._icon_match_events: List[event_listeners.IconMatchEvent] = []
    self._view_hierarchy_events: List[event_listeners.ViewHierarchyEvent] = []
    self._log_events: List[event_listeners.LogEvent] = []
    self._log_filters: Set[str] = set()
    self._response_events: List[event_listeners.ResponseEvent] = []

    for evt_s in task.event_sources:
      self._parse_event_source(evt_s)

    # event slots
    self._score_event: event_listeners.EventSlot[Any, Any, float] =\
      self._parse_event_listeners(task.event_slots.score_listener, wrap=float)\
        if task.HasField("event_slots") and task.event_slots.HasField("score_listener")\
        else event_listeners.EmptyEvent()
    self._reward_event: event_listeners.EventSlot[Any, Any, float] =\
      self._parse_event_listeners(task.event_slots.reward_listener, wrap=float,
          update=operator.add)\
        if task.HasField("event_slots") and task.event_slots.HasField("reward_listener")\
        else event_listeners.EmptyEvent()
    self._episode_end_event: event_listeners.EventSlot[Any, Any, Any] =\
      self._parse_event_listeners(task.event_slots.episode_end_listener)\
        if task.HasField("event_slots") and task.event_slots.HasField("episode_end_listener")\
        else event_listeners.EmptyEvent()

    def _update_dict(buffer_size_limit: int,
        dict1: Dict[str, str],
        dict2: Dict[str, str]) -> Dict[str, str]:
      #  function `_update_dict` {{{ # 
      for k in dict2:
        if k in dict1:
          dict1[k] += dict2[k]
        else:
          dict1[k] = dict2[k]
        if len(dict1[k])>buffer_size_limit:
          dict1[k] = dict1[k][-buffer_size_limit:]
      return dict1
      #  }}} function `_update_dict` # 
    self._extra_event: event_listeners.EventSlot[Any, Any, Dict[str, str]] =\
      self._parse_event_listeners(task.event_slots.extra_listener,
          update=functools.partial(_update_dict, self._extras_max_buffer_size))\
        if task.HasField("event_slots") and task.event_slots.HasField("extra_listener")\
        else event_listeners.EmptyEvent()

    def _parse_json(extra: Union[Dict[str, Any], str]) -> Dict[str, Any]:
      #  function `_parse_json` {{{ # 
      if isinstance(extra, dict):
        return extra

      try:
        extra = dict(json.loads(extra))
      except ValueError:
        logging.error('JSON string could not be parsed: %s', extra)
        extra = {}
      return extra
      #  }}} function `_parse_json` # 
    self._json_extra_event: event_listeners.EventSlot[Any,
        Union[Dict[str, Any], str],
        Dict[str, str]] =\
      self._parse_event_listeners(task.event_slots.json_extra_listener,
          wrap=_parse_json,
          update=functools.partial(_update_dict, self._extras_max_buffer_size))\
        if task.HasField("event_slots") and task.event_slots.HasField("json_extra_listener")\
        else event_listeners.EmptyEvent()

    def _update_list(buffer_size_limit: int, instruction1: List[str], instruction2: List[str]) -> List[str]:
      instruction1 += instruction2
      return instruction1
    self._instruction_event: event_listeners.EventSlot[Any,
        Union[List[str], str],
        List[str]] =\
      self._parse_event_listeners(task.event_slots.instruction_listener,
          wrap=(lambda instrct: instrct if isinstance(instrct, list) else [instrct]),
          update=functools.partial(_update_list, self._extras_max_buffer_size))\
        if task.HasField("event_slots") and task.event_slots.HasField("instruction_listener")\
        else event_listeners.EmptyEvent()

    del self._events_with_id
    del self._events_in_need
    del self._events_in_source
    #  }}} Event Infrastructures # 

    self._text_model: TextModel = text_model
    self._icon_model: IconModel = icon_model

    # Initialize internal state
    self._task_start_time = None
    self._episode_steps = 0
    self._bad_state_counter = 0
    self._is_bad_episode = False
    self._setup_flag: bool = False

    self._latest_values = {
        #'reward': 0.0,
        'score': 0.0,
        "player_exited": False,
        #'extra': {},
        #'episode_end': False,
    }

    #logging.info('Task config: %s', self._task)
    #  }}} method `__init__` # 

  # zdy
  def _parse_event_source(self, event_source: task_pb2.EventSource):
    #  method `_parse_event_source` {{{ # 
    """
    event_source - task_pb2.EventSource
    """

    def _rect_to_list(rect: task_pb2.EventSource.Rect) -> List[float]:
      #  function `_rect_to_list` {{{ # 
      """
      rect - task_pb2.EventSource.Rect

      return list of four floats
      """

      return [rect.x0, rect.y0, rect.x1, rect.y1]
      #  }}} function `_rect_to_list` # 

    #  Text Events {{{ # 
    if event_source.HasField("text_recognize"):
      event = event_listeners.TextEvent(event_source.text_recognize.expect,
          _rect_to_list(event_source.text_recognize.rect), needs_detection=False,
          repeatability=event_source.repeatability)
      self._text_events.append(event)
    elif event_source.HasField("text_detect"):
      event = event_listeners.TextEvent(event_source.text_detect.expect,
          _rect_to_list(event_source.text_detect.rect), needs_detection=True,
          repeatability=event_source.repeatability)
      self._text_events.append(event)
    #  }}} Text Events # 
    #  Icon Events {{{ # 
    elif event_source.HasField("icon_recognize"):
      event = event_listeners.IconRecogEvent(getattr(event_source.icon_recognize, "class"),
          _rect_to_list(event_source.icon_recognize.rect), needs_detection=False,
          repeatability=event_source.repeatability)
      self._icon_events.append(event)
    elif event_source.HasField("icon_detect"):
      event = event_listeners.IconRecogEvent(getattr(event_source.icon_detect, "class"),
          _rect_to_list(event_source.icon_detect.rect), needs_detection=True,
          repeatability=event_source.repeatability)
      self._icon_events.append(event)
    #  }}} Icon Events # 
    #  Icon Match Events {{{ # 
    elif event_source.HasField("icon_match"):
      event = event_listeners.IconMatchEvent(event_source.icon_match.path,
          _rect_to_list(event_source.icon_match.rect), needs_detection=False,
          repeatability=event_source.repeatability)
      self._icon_match_events.append(event)
    elif event_source.HasField("icon_detect_match"):
      event = event_listeners.IconMatchEvent(event_source.icon_detect_match.path,
          _rect_to_list(event_source.icon_detect_match.rect), needs_detection=True,
          repeatability=event_source.repeatability)
      self._icon_match_events.append(event)
    #  }}} Icon Match Events # 
    #  Other Events {{{ # 
    elif event_source.HasField("view_hierarchy_event"):
      properties = []
      for prpt in event_source.view_hierarchy_event.properties:
        #if not prpt.HasField("property_value"):
          #property_ = event_listeners.ViewHierarchyEvent.PureProperty(prpt.property_name)
        if prpt.HasField("pattern"):
          property_ = event_listeners.ViewHierarchyEvent.StringProperty(prpt.property_name, prpt.pattern)
        else:
          if prpt.sign==task_pb2.Event.ViewHierarchyEvent.Sign.EQ:
            comparator = operator.eq
          elif prpt.sign==task_pb2.Event.ViewHierarchyEvent.Sign.LE:
            comparator = operator.le
          elif prpt.sign==task_pb2.Event.ViewHierarchyEvent.Sign.LT:
            comparator = operator.lt
          elif prpt.sign==task_pb2.Event.ViewHierarchyEvent.Sign.GE:
            comparator = operator.ge
          elif prpt.sign==task_pb2.Event.ViewHierarchyEvent.Sign.GT:
            comparator = operator.gt
          else:
            comparator = operator.ne
          property_ = event_listeners.ViewHierarchyEvent.ScalarProperty(prpt.property_name,
              comparator, getattr(prpt, prpt.WhichOneof("property_value")))
        properties.append(property_)
      event = event_listeners.ViewHierarchyEvent(event_source.view_hierarchy_event.view_hierarchy_path,
          properties, repeatability=event_source.repeatability)
      self._view_hierarchy_events.append(event)
    elif event_source.HasField("log_event"):
      self._log_filters |= set(event_source.log_event.filters)
      event = event_listeners.LogEvent(event_source.log_event.filters, event_source.log_event.pattern,
          repeatability=event_source.repeatability)
      self._log_events.append(event)
    elif event_source.HasField("response_event"):
      event = event_listeners.ResponseEvent( event_source.response_event.pattern
                                           , repeatability=event_source.repeatability
                                           )
      self._response_events.append(event)
    #  }}} Other Events # 

    #  Handle the id {{{ # 
    if event_source.id!=0:
      self._events_with_id[event_source.id] = event
    #  }}} Handle the id # 
    #  }}} method `_parse_event_source` # 

  # zdy
  def _parse_event_listeners(self, event_definition: task_pb2.EventSlot,
      #cast: Optional[Callable[[event_listeners.V], event_listeners.C]] = None,
      wrap: Optional[Callable[[event_listeners.T], event_listeners.W]] = None,
      update: Optional[Callable[[event_listeners.W, event_listeners.W],
        event_listeners.W]] = None)\
            -> event_listeners.EventSlot[event_listeners.V,
                event_listeners.T,
                event_listeners.W]:
    #  method `_parse_event_listeners` {{{ # 
    """
    Parses the definition of event slot into an inner EventSlot instance.

    event_definition - task_pb2.EventSlot
    #cast - callable accepting the verified type returning the cast type or
      #None
    wrap - callable accepting the transformed type returning the wrapped
      type or None
    update - callable accepting
      + the wrapped type
      + the wrapped type
      and returning the wrapped type

    return event_listeners.EventSlot
    """

    transformation = event_definition.transformation if len(event_definition.transformation)>0\
        else ["y = x"]

    if len(event_definition.events)==0:
      return event_listeners.EmptyEvent()

    #  Combined Events {{{ # 
    late_source_ids = []

    # handle sub-events
    sub_events = []
    for i, s_evt_dfnt in enumerate(event_definition.events):
      if s_evt_dfnt.HasField("id"):
        if s_evt_dfnt.id in self._events_with_id:
          sub_event = self._events_with_id[s_evt_dfnt.id]
        else:
          sub_event = None
          late_source_ids.append((i, s_evt_dfnt.id))
      elif s_evt_dfnt.HasField("event"):
        sub_event = self._parse_event_listeners(s_evt_dfnt.event)
      sub_events.append(sub_event)

    # create event object
    if event_definition.type==task_pb2.EventSlot.Type.SINGLE:
      event = event_listeners.DefaultEvent([sub_events[0]],
          transformation=transformation, wrap=wrap, update=update)
      if len(late_source_ids)>0:
        late_source_ids = [late_source_ids[0]]
    elif event_definition.type==task_pb2.EventSlot.Type.OR:
      event = event_listeners.Or(sub_events, transformation, wrap, update)
    elif event_definition.type==task_pb2.EventSlot.Type.AND:
      event = event_listeners.And(sub_events, transformation, wrap)

    # handle late sources
    if len(late_source_ids)>0:
      for i, id_ in late_source_ids:
        if id_ not in self._events_in_source:
          self._events_in_source[id_] = []
        self._events_in_source[id_].append((i, event))
    #  }}} Combined Events # 

    #  Handle the prerequisites {{{ # 
    if event_definition.id!=0:
      event_id = event_definition.id
      self._events_with_id[event_id] = event

      if event_id in self._events_in_need:
        for evt in self._events_in_need[event_id]:
          evt.add_prerequisites(event)
        del self._events_in_need[event_id]

      if event_id in self._events_in_source:
        for i, evt in self._events_in_source[event_id]:
          evt.replace_sources(i, event)
        del self._events_in_source[event_id]
    if len(event_definition.prerequisite)>0:
      for evt_id in event_definition.prerequisite:
        if evt_id in self._events_with_id:
          event.add_prerequisites(self._events_with_id[evt_id])
        else:
          if evt_id not in self._events_in_need:
            self._events_in_need[evt_id] = []
          self._events_in_need[evt_id].append(event)
    #  }}} Handle the prerequisites # 

    return event
    #  }}} method `_parse_event_listeners` # 

  #  Properties {{{ # 
  def task(self) -> task_pb2.Task:
    return self._task
  def command(self) -> List[str]:
    """
    return list of str
    """

    return list(self._task.command)
  def nb_tokens(self) -> int:
    """
    return int
    """

    return len(self._vocabulary)
  def vocabulary(self) -> List[str]:
    """
    return list of str
    """

    return self._vocabulary
  #  }}} Properties # 

  def increment_steps(self):
    self._episode_steps += 1

  def log_dict(self) -> Dict[str, Any]:
    log_dict = copy.deepcopy(self._log_dict)
    log_dict.update(self._setup_step_interpreter.log_dict())
    return log_dict

  #  Episode Management {{{ # 
  def _reset_counters(self):
    #  method `_reset_counters` {{{ # 
    """Reset counters at the end of an RL episode."""

    if not self._is_bad_episode:
      self._bad_state_counter = 0
    self._is_bad_episode = False

    self._episode_steps = 0
    self._task_start_time = datetime.datetime.now()
    with self._lock:
      self._latest_values = {
          #'reward': 0.0,
          'score': 0.0,
          "player_exited": False,
          #'extra': {},
          #'episode_end': False,
      }

      # clear event sources
      self.clear_events()

      # reset event slots
      self._score_event.reset()
      self._reward_event.reset()
      self._episode_end_event.reset()
      self._extra_event.reset()
      self._json_extra_event.reset()
      self._instruction_event.reset()
    #  }}} method `_reset_counters` # 

  def setup_task(self,
                 adb_controller: adb_control.AdbController,
                 #emulator_stub: emulator_controller_pb2_grpc.EmulatorControllerStub, # zdy
                 #image_format: emulator_controller_pb2.ImageFormat, # zdy
                 log_stream: log_stream_lib.LogStream,
                 ) -> None:
    """Starts the given task along with all relevant processes."""

    logging.info("#Text Events: {:d}".format(len(self._text_events)))
    logging.info("#Icon Events: {:d}".format(len(self._icon_events)))
    logging.info("#Icon Match Events: {:d}".format(len(self._icon_match_events)))
    logging.info("#VH Events: {:d}".format(len(self._view_hierarchy_events)))
    logging.info("#Log Events: {:d}".format(len(self._log_events)))
    logging.info("#Response Events: {:d}".format(len(self._response_events)))

    self._adb_controller: adb_control.AdbController = adb_controller
    #self._emulator_stub = emulator_stub
    #self._image_format = image_format
    self._start_logcat_thread(log_stream=log_stream)
    self._start_setup_step_interpreter()
    self._setup_step_interpreter.interpret(self._task.setup_steps)

    self._setup_flag = True

  def reset_task(self) -> None:
    """Resets a task at the end of an RL episode."""

    self.pause_task()
    self._setup_step_interpreter.interpret(self._task.reset_steps)
    self._resume_task()
    self._reset_counters()

  def pause_task(self) -> None:
    self._stop_dumpsys_thread()
    self._stop_screen_analyzer_thread() # zdy

  def _resume_task(self) -> None:
    self._start_dumpsys_thread()
    self._start_screen_analyzer_thread() # zdy

  def setup_flag(self) -> bool:
    return self._setup_flag
  def clear_setup_flag(self):
    self._setup_flag = False
  #  }}} Episode Management # 

  #  Interaction Methods {{{ # 
  def receive_response(self, response: str):
    for l in response.splitlines():
      for evt in self._response_events:
        evt.set(l)

  def send_token(self, token_id: int):
    #  method `send_token` {{{ # 
    """
    token_id - int
    """

    token = self._vocabulary[token_id]

    if self._special_token_pattern.match(token):
      return

    logging.info("\x1b[31;42mINPUT: \x1b[31m{:}\x1b[0m".format(self._vocabulary[token_id]))

    if self._non_start_mark_first:
      is_non_start_token = token.startswith(self._non_start_token_mark)
      is_start_token = not is_non_start_token and token.startswith(self._start_token_mark)
    else:
      is_start_token = token.startswith(self._start_token_mark)
      is_non_start_token = not is_start_token and token.startswith(self._non_start_token_mark)

    if is_start_token:
      token = token[len(self._start_token_mark):]
    elif is_non_start_token:
      token = token[len(self._non_start_token_mark):]

    if not token.isascii():
      return
    if is_non_start_token:
      self._adb_controller.input_key("KEYCODE_DEL")

    for n_prtb, gr in itertools.groupby( token
                                       , key=( lambda ch:
                                                 ch in TaskManager.non_pritable_char_mapping
                                             )
                                       ):
      if n_prtb:
        for ch in gr:
          self._adb_controller.input_key(TaskManager.non_pritable_char_mapping[ch])
      else:
        self._adb_controller.input_text( "\""\
                                       + "".join(gr)\
                                          .replace("\\", "\\\\")\
                                          .replace("\"", "\\\"")\
                                       + "\""
                                       )
    self._adb_controller.input_text("\" \"")
    #  }}} method `send_token` # 

  def get_current_reward(self, vh: bool) -> Tuple[ float
                                                 , Optional[lxml.etree.Element]
                                                 ]:
    #  method `get_current_reward` {{{ # 
    """
    Returns total reward accumulated since the last step.

    Args:
        vh (bool): if view hierarchy should be obtained

    Returns:
        float: reward
        #str: view hierarchy serialization
        Optional[lxml.etree.Element]: view hierarchy
    """

    #with self._lock:
      #reward = self._latest_values['reward']
      #self._latest_values['reward'] = 0.0

    # zdy
    if self._score_event.is_set():
      score = self._score_event.get()[0]
      reward = score - self._latest_values["score"]
      self._latest_values["score"] = score
      #self._score_event.clear()
    else:
      reward = 0

    reward += self._reward_event.get()[0] if self._reward_event.is_set() else 0 # zdy
    #self._reward_event.clear() # zdy

    if vh:
      view_hierarchy = self._adb_controller.get_view_hierarchy()
    else:
      view_hierarchy = None
    #view_hierarchy = lxml.etree.tostring(view_hierarchy, encoding="unicode")\
        #if view_hierarchy is not None\
        #else ""

    return reward, view_hierarchy
    #  }}} method `get_current_reward` # 

  def get_current_extras(self) -> Dict[str, Any]:
    #  method `get_current_extras` {{{ # 
    """Returns task extras accumulated since the last step."""

    #with self._lock:
      # zdy
    extras = self._extra_event.get()[0] if self._extra_event.is_set() else {}
    #self._extra_event.clear()
    #extras = {k: list(map(ast.literal_eval, vals)) for k, vals in extras.items()}
    extras = extras.copy()

    if self._json_extra_event.is_set():
      json_extras = self._json_extra_event.get()[0]
      #self._json_extra_event.clear()
    else:
      json_extras = {}

    if not isinstance(json_extras, dict):
      try:
        json_extras = json.loads(json_extras)
      except ValueError:
        logging.error('JSON string could not be parsed: %s', json_extras)
        json_extras = {}

    for k in json_extras:
      if k in extras:
        extras[k] += json_extras[k]
      else:
        extras[k] = json_extras[k]

    #for name, values in self._latest_values['extra'].items(): # zdy
    for name, values in extras.items(): # zdy
      extras[name] = np.stack(values)
    #self._latest_values['extra'] = {} # zdy
    return extras
    #  }}} method `get_current_extras` # 

  def get_current_instructions(self) -> List[str]:
    #  function `get_current_instructions` {{{ # 
    """
    return list of str
    """

    #with self._lock:
    instructions = self._instruction_event.get()[0] if self._instruction_event.is_set() else []
    #self._instruction_event.clear()

    if not isinstance(instructions, list):
      instructions = [instructions]
    else:
      instructions = instructions.copy()

    return instructions
    #  }}} function `get_current_instructions` # 

  def check_if_episode_ended(self, with_view_hierarchy: bool) -> bool:
    #  method `check_if_episode_ended` {{{ # 
    """Determines whether the episode should be terminated and reset."""

    # Check if player existed the task
    #if self._check_player_exited():
    if self._latest_values["player_exited"]:
      self._log_dict['reset_count_player_exited'] += 1
      logging.warning('Player exited the game. Ending episode.')
      logging.info('************* END OF EPISODE *************')
      return True

    # Check if episode has ended
    #with self._lock:
      #if self._latest_values['episode_end']: # zdy
    if self._episode_end_event.is_set(): # zdy
      self._log_dict['reset_count_episode_end'] += 1
      logging.info('End of episode from logcat! Ending episode.')
      logging.info('************* END OF EPISODE *************')
      return True

    # Check if step limit or time limit has been reached
    if self._task.max_num_steps > 0:
      if self._episode_steps > self._task.max_num_steps:
        self._log_dict['reset_count_max_duration_reached'] += 1
        logging.info('Maximum task duration (steps) reached. Ending episode.')
        logging.info('************* END OF EPISODE *************')
        return True

    if not with_view_hierarchy and self._task.max_duration_sec > 0.0:
      task_duration = datetime.datetime.now() - self._task_start_time
      max_duration_sec = self._task.max_duration_sec
      if task_duration > datetime.timedelta(seconds=int(max_duration_sec)):
        self._log_dict['reset_count_max_duration_reached'] += 1
        logging.info('Maximum task duration (sec) reached. Ending episode.')
        logging.info('************* END OF EPISODE *************')
        return True

    return False
    #  }}} method `check_if_episode_ended` # 

  def snapshot_events(self, screen: np.ndarray):
    #  method `snapshot_events` {{{ # 
    """
    Args:
      screen: array of uint8 with shape (height, width, 3)
    """

    # zdy
    self._run_screen_analyzer(np.transpose(screen, axes=(2, 0, 1)))
    try:
      self._run_dumpsys()
    except errors.NotAllowedError:
      self._latest_values["player_exited"] = True

    with self._lock:
      for evt in itertools.chain( self._text_events
                                , self._icon_events
                                , self._icon_match_events
                                , self._view_hierarchy_events
                                , self._log_events
                                , self._response_events
                                ):
        evt.snapshot()
    #  }}} method `snapshot_events` # 

  def clear_events(self):
    #  method `clear_events` {{{ # 
    #with self._lock:
    for evt in itertools.chain( self._text_events
                              , self._icon_events
                              , self._icon_match_events
                              , self._view_hierarchy_events
                              , self._log_events
                              , self._response_events
                              ):
      evt.clear()
    #  }}} method `clear_events` # 

  #  Deprecated `_check_player_exited` method {{{ # 
  # zdy
  #def _check_player_exited(self) -> bool:
    #"""Returns whether the player has exited the game."""
    #try:
      #self._check_player_exited_impl()
      #return False
    #except errors.NotAllowedError:
      #return True
  #  }}} Deprecated `_check_player_exited` method # 

  #def _check_player_exited_impl(self):
  def _run_dumpsys(self): # zdy
    #  method `_run_dumpsys` {{{ # 
    """Raises an error if the OS is not in an allowed state."""

    if not hasattr(self, '_dumpsys_thread'):
      return

    self._dumpsys_thread.write(
        dumpsys_thread.DumpsysThread.Signal.FETCH_DUMPSYS)

    try:
      v = self._dumpsys_thread.read(block=False) # ZDY_COMMENT: TODO: whether block or not and a probable proper timeout value could be tested.
      if v == dumpsys_thread.DumpsysThread.Signal.USER_EXITED_ACTIVITY:
        self._increment_bad_state()
        raise errors.PlayerExitedActivityError()
      elif v == dumpsys_thread.DumpsysThread.Signal.USER_EXITED_VIEW_HIERARCHY:
        self._increment_bad_state()
        raise errors.PlayerExitedViewHierarchyError()
    except queue.Empty:
      pass  # Don't block here, just ignore if we have nothing.
    #  }}} method `_run_dumpsys` # 

  # zdy
  def _run_screen_analyzer(self, screen: np.ndarray):
    #  method `_run_screen_analyzer` {{{ # 
    """
    screen - array of uint8 with shape (3, height, width)
    """

    if not hasattr(self, "_screen_analyzer_thread"):
      return

    self._screen_analyzer_thread.write(
      screen_analyzer_thread.ScreenAnalyzerThread.Signal.CHECK_SCREEN)
    self._screen_analyzer_thread.write(screen)

    try:
      status = self._screen_analyzer_thread.read(block=True, timeout=0.05) # TODO: maybe a better timeout value could be tuned to
      if status==screen_analyzer_thread.ScreenAnalyzerThread.Signal.CHECK_ERROR:
        logging.error("Screen Analyzer Error!")
      elif status==screen_analyzer_thread.ScreenAnalyzerThread.Signal.DID_NOT_CHECK:
        logging.info("Screen Analyzer did not check, maybe owing to the kill signal.")
    except queue.Empty:
      logging.warning("Screen Analyzer exceeds time limit!")
    #  }}} method `_run_screen_analyzer` # 
  #  }}} Interaction Methods # 

  #  Methods to Start and Stop the Assistant Threads {{{ # 
  def _start_setup_step_interpreter(self):
    self._setup_step_interpreter = setup_step_interpreter.SetupStepInterpreter(
        adb_controller=self._adb_controller,
        logcat=self._logcat_thread)

  def _start_logcat_thread(self, log_stream: log_stream_lib.LogStream):
    #self._logcat_thread = logcat_thread.LogcatThread(
        #log_stream=log_stream,
        #log_parsing_config=self._task.log_parsing_config)
    # zdy
    self._logcat_thread = logcat_thread.LogcatThread(
        log_stream=log_stream,
        log_filter=self._log_filters,
        lock=self._lock)

    #for event_listener in self._logcat_listeners():
    for event_listener in self._log_events: # zdy
      self._logcat_thread.add_event_listener(event_listener)

  def _start_dumpsys_thread(self):
    app_screen_checker_ = app_screen_checker.AppScreenChecker(
        self._adb_controller, self._task.expected_app_screen)
    app_screen_checker_.add_event_listeners(*self._view_hierarchy_events)
    self._dumpsys_thread = dumpsys_thread.DumpsysThread(
        app_screen_checker=app_screen_checker_,
        check_frequency=self._dumpsys_check_frequency,
        max_failed_current_activity=self._max_failed_current_activity,
        lock=self._lock,
        block_input=True,
        block_output=True)

  # zdy
  def _start_screen_analyzer_thread(self):
    #  method `_start_screen_analyzer_thread` {{{ # 
    # ZDY_COMMENT: TODO: instantiate models according to requirements

    #text_detector = naive_functions.naive_text_detector
    #text_recognizer = naive_functions.naive_text_recognizer
    #icon_detector = naive_functions.naive_icon_detector
    #icon_recognizer = naive_functions.naive_icon_recognizer
    #icon_matcher = naive_functions.naive_icon_matcher
    text_detector = self._text_model.text_detector
    text_recognizer = self._text_model.text_recognizer
    icon_detector = self._icon_model.icon_detector
    icon_recognizer = self._icon_model.icon_recognizer
    icon_matcher = self._icon_model.icon_matcher

    self._screen_analyzer_thread: screen_analyzer_thread.ScreenAnalyzerThread =\
        screen_analyzer_thread.ScreenAnalyzerThread(
          text_detector, text_recognizer,
          icon_detector, icon_recognizer, icon_matcher,
          #self._emulator_stub, self._image_format,
          lock=self._lock,
          block_input=True, block_output=True)
    self._screen_analyzer_thread.add_text_event_listeners(*self._text_events)
    self._screen_analyzer_thread.add_icon_event_listeners(*self._icon_events)
    self._screen_analyzer_thread.add_icon_match_event_listeners(*self._icon_match_events)

    #  }}} method `_start_screen_analyzer_thread` # 

  def _stop_logcat_thread(self):
    if hasattr(self, '_logcat_thread'):
      self._logcat_thread.kill()

  def _stop_dumpsys_thread(self):
    if hasattr(self, '_dumpsys_thread'):
      self._dumpsys_thread.kill()

  # zdy
  def _stop_screen_analyzer_thread(self):
    if hasattr(self, "_screen_analyzer_thread"):
      self._screen_analyzer_thread.kill()
  #  }}} Methods to Start and Stop the Assistant Threads # 

  def _increment_bad_state(self) -> None:
    """Increments the bad state counter.

    Bad states are errors that shouldn't happen and that trigger an
    episode reset. If enough bad states have been seen consecutively,
    we restart the simulation in the hope of returning the simulation
    to a good state.
    """
    logging.warning('Bad state detected.')
    if self._max_bad_states:
      self._is_bad_episode = True
      self._bad_state_counter += 1
      logging.warning('Bad state counter: %d.', self._bad_state_counter)
      if self._bad_state_counter >= self._max_bad_states:
        logging.error('Too many consecutive bad states. Restarting simulator.')
        self._log_dict['restart_count_max_bad_states'] += 1
        self._should_restart = True
    else:
      logging.warning('Max bad states not set, bad states will be ignored.')

  # ZDY_COMMENT: I think it is time to obsolete this method
  def _logcat_listeners(self):
    #  method `_logcat_listeners` {{{ # 
    """Creates list of EventListeners for logcat thread."""

    # Defaults to 'a^' since that regex matches no string by definition.
    regexps = self._task.log_parsing_config.log_regexps
    listeners = []

    # Reward listeners
    def _reward_handler(event, match):
      del event
      reward = float(match.group(1))
      with self._lock:
        self._latest_values['reward'] += reward
        # ZDY_COMMENT: `_latest_values['reward']` is cleared during revoking `get_current_reward`

    for regexp in regexps.reward:
      listeners.append(logcat_thread.EventListener(
          regexp=re.compile(regexp or 'a^'),
          handler_fn=_reward_handler))

    # RewardEvent listeners
    # ZDY_COMMENT: RewardEvent returns a constant reward regardless of the log detail
    for reward_event in regexps.reward_event:

      def get_reward_event_handler(reward):
        def _reward_event_handler(event, match):
          del event, match
          with self._lock:
            self._latest_values['reward'] += reward
        return _reward_event_handler

      listeners.append(logcat_thread.EventListener(
          regexp=re.compile(reward_event.event or 'a^'),
          handler_fn=get_reward_event_handler(reward_event.reward)))

    # Score listener
    # ZDY_COMMENT: the increment of the score constructs the reward
    def _score_handler(event, match):
      del event
      current_score = float(match.group(1))
      with self._lock:
        current_reward = current_score - self._latest_values['score']
        self._latest_values['score'] = current_score
        self._latest_values['reward'] += current_reward

    listeners.append(logcat_thread.EventListener(
        regexp=re.compile(regexps.score or 'a^'),
        handler_fn=_score_handler))

    # Episode end listeners
    def _episode_end_handler(event, match):
      del event, match
      with self._lock:
        self._latest_values['episode_end'] = True

    for regexp in regexps.episode_end:
      listeners.append(logcat_thread.EventListener(
          regexp=re.compile(regexp or 'a^'),
          handler_fn=_episode_end_handler))

    # Extra listeners
    def _extras_handler(event, match):
      del event
      extra_name = match.group('name')
      extra = match.group('extra')
      if extra:
        try:
          extra = ast.literal_eval(extra)
          # ZDY_BOOKMARK: What's exatly in this
          # ZDY_COMMENT: oh, got it, maybe a list of basic type elements
        # Except all to avoid unnecessary crashes, only log error.
        except Exception:  # pylint: disable=broad-except
          logging.exception('Could not parse extra: %s', extra)
      else:
        # No extra value provided for boolean extra. Setting value to True.
        extra = 1
      _process_extra(extra_name, extra)

    for regexp in regexps.extra:
      listeners.append(logcat_thread.EventListener(
          regexp=re.compile(regexp or 'a^'),
          handler_fn=_extras_handler))

    # JSON extra listeners
    def _json_extras_handler(event, match):
      del event
      extra_data = match.group('json_extra')
      try:
        extra = dict(json.loads(extra_data))
      except ValueError:
        logging.error('JSON string could not be parsed: %s', extra_data)
        return
      for extra_name, extra_value in extra.items():
        _process_extra(extra_name, extra_value)

    for regexp in regexps.json_extra:
      listeners.append(logcat_thread.EventListener(
          regexp=re.compile(regexp or 'a^'),
          handler_fn=_json_extras_handler))

    def _process_extra(extra_name, extra):
      extra = np.array(extra)
      with self._lock:
        latest_extras = self._latest_values['extra']
        if extra_name in latest_extras:
          # If latest extra is not flushed, append.
          if len(latest_extras[extra_name]) >= self._extras_max_buffer_size:
            latest_extras[extra_name].pop(0)
          latest_extras[extra_name].append(extra)
        else:
          latest_extras[extra_name] = [extra]
        self._latest_values['extra'] = latest_extras
    # ZDY_COMMENT: multiple latest extras are buffered, maybe for the purpose to preserve the information of dynamics

    return listeners
    #  }}} method `_logcat_listeners` # 

  def close(self):
    if hasattr(self, '_logcat_thread'):
      self._logcat_thread.kill()
    if hasattr(self, '_dumpsys_thread'):
      self._dumpsys_thread.kill()
    # zdy
    if hasattr(self, "_screen_analyzer_thread"):
      self._screen_analyzer_thread.kill()
    if hasattr(self, "_adb_controller"):
      self._adb_controller.close()



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Definitions of exceptions used by AndroidEnv."""


class ReadObservationError(Exception):
  """When the environment is unable to obtain an observation from a simulator."""


class ObservationDecodingError(ReadObservationError):
  """When the environment is unable to decode the observation from a simulator."""


class PipeTimedOutError(ReadObservationError):
  """When the environment waited for too long for part of an observation."""


class CoordinatorError(Exception):
  """Error raised by the Coordinator."""


class TooManyRestartsError(CoordinatorError):
  """The number of restarts has exceeded _MAX_RESTART_TRIES."""


class NotAllowedError(Exception):
  """When the player does something that outside of the task scope."""


class PlayerExitedActivityError(NotAllowedError):
  """When the player quits the current Android activity."""


class PlayerExitedViewHierarchyError(NotAllowedError):
  """When the player quits the current Android app screen."""


class AdbControllerError(Exception):
  """Errors that can be raised by ADBController."""


class AdbControllerShellInitError(AdbControllerError):
  """Raised when an error occurred when initializing ADB shell."""


class AdbControllerPexpectError(AdbControllerError):
  """Raise when a problem with pexpect communication occurs."""


class AdbControllerDeviceTimeoutError(AdbControllerError):
  """Raised when a device takes too long to respond."""


class AdbControllerConnectionError(AdbControllerError):
  """Error connecting to tcpip address."""


class AdbControllerInputDeviceNotFoundError(AdbControllerError):
  """Raised when an input device handle could not be found."""


class AdbControllerScreenPinningError(AdbControllerError):
  """Raised when screen pinning failed."""


class SimulatorCrashError(Exception):
  """Raised when an AndroidSimulator crashed."""


class ConsoleConnectionError(Exception):
  """Raised when cannot connect to the emulator console."""


class SendActionError(Exception):
  """Raised when action couldn't be sent successfully."""


class StepCommandError(Exception):
  """Raised when setup step interpreter cannot process a command."""


class WaitForAppScreenError(StepCommandError):
  """Raised when the wait_for_app_screen success check is not met."""


class CheckInstallError(StepCommandError):
  """Raised when the check_install success check is not met."""


class WaitForMessageError(StepCommandError):
  """Raised when the wait_for_message success check is not met."""



# coding=utf-8
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for android_env.components.thread_function."""

from absl.testing import absltest
from android_env.components import thread_function


class ThreadFunctionTest(absltest.TestCase):

  def test_hello_world(self):

    class HelloFunction(thread_function.ThreadFunction):

      def main(self):
        v = self._read_value()
        if v:
          self._write_value('hello')
        else:
          self._write_value('world')

    hello_fn = HelloFunction(block_input=True, block_output=True, name='Hello')
    hello_fn.write(True, block=True)
    self.assertEqual(hello_fn.read(block=True), 'hello')
    hello_fn.write(False, block=True)
    self.assertEqual(hello_fn.read(block=True), 'world')
    hello_fn.kill()  # Not strictly necessary, but it's good hygiene.

  def test_class_with_state(self):
    """Derived class with internal state to change the thread output."""

    class HaveState(thread_function.ThreadFunction):

      def __init__(self, value, *args, **kwargs):
        self._value = value
        super().__init__(*args, **kwargs)

      def main(self):
        v = self._read_value()
        if v:
          self._write_value(v + self._value)

    have_state_fn = HaveState(
        value=123, block_input=True, block_output=True, name='HaveState')
    have_state_fn.write(3, block=True)
    self.assertEqual(have_state_fn.read(block=True), 126)
    have_state_fn.write(9, block=True)
    self.assertEqual(have_state_fn.read(block=True), 132)
    have_state_fn.kill()


if __name__ == '__main__':
  absltest.main()



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base specs for AndroidEnv."""

from typing import Dict

from android_env.components import action_type
from android_env.proto import task_pb2
import dm_env
from dm_env import specs
import numpy as np


_PROTO_DTYPE_TO_NUMPY_DTYPE = {
    task_pb2.ArraySpec.DataType.FLOAT: np.float32,
    task_pb2.ArraySpec.DataType.DOUBLE: np.float64,
    task_pb2.ArraySpec.DataType.INT8: np.int8,
    task_pb2.ArraySpec.DataType.INT16: np.int16,
    task_pb2.ArraySpec.DataType.INT32: np.int32,
    task_pb2.ArraySpec.DataType.INT64: np.int64,
    task_pb2.ArraySpec.DataType.UINT8: np.uint8,
    task_pb2.ArraySpec.DataType.UINT16: np.uint16,
    task_pb2.ArraySpec.DataType.UINT32: np.uint32,
    task_pb2.ArraySpec.DataType.UINT64: np.uint64,
    task_pb2.ArraySpec.DataType.BOOL: np.bool_,
    task_pb2.ArraySpec.DataType.STRING_U1: np.dtype(('U1')),
    task_pb2.ArraySpec.DataType.STRING_U16: np.dtype(('<U16')),
    task_pb2.ArraySpec.DataType.STRING_U25: np.dtype(('<U25')),
    task_pb2.ArraySpec.DataType.STRING_U250: np.dtype(('<U250')),
}


def base_action_spec(num_tokens: int, num_fingers: int = 1) -> Dict[str, specs.Array]:
  """Default action spec for AndroidEnv.

  Args:
    num_tokens: Number of candidate tokens
    num_fingers: Number of virtual fingers of the agent.
  Returns:
    A dict of action specs, each item corresponding to a virtual finger.
    action_type: An integer of type ActionType: TOUCH=0, LIFT=1, REPEAT=2, TEXT=3
    touch_position: Position [x, y] of the touch action, where x, y are float
      values between 0.0 and 1.0 corresponding to the relative position on the
      screen. IGNORED when (action_type != ActionType.TOUCH).
    input_token: An integer as the input token id.
    action_type_i: Action type for additional fingers (i>1).
    touch_position_i: Touch position for additional fingers (i>1).
    response: text
  """

  nb_types = len(action_type.ActionType)-1 if num_tokens==0 else len(action_type.ActionType)
  action_spec = { 'action_type':
                    specs.DiscreteArray( num_values=nb_types
                                       , name='action_type'
                                       )
                , 'touch_position':
                    specs.BoundedArray( shape=(2,)
                                      , dtype=np.float32
                                      , minimum=[0.0, 0.0]
                                      , maximum=[1.0, 1.0]
                                      , name='touch_position'
                                      )
                , 'response': specs.StringArray( shape=()
                                               , name="response"
                                               )
                }

  for i in range(2, num_fingers + 1):
    action_spec.update({
        f'action_type_{i}':
            specs.DiscreteArray(
                num_values=len(action_type.ActionType)-1,
                name=f'action_type_{i}'),
        f'touch_position_{i}':
            specs.BoundedArray(
                shape=(2,),
                dtype=np.float32,
                minimum=[0.0, 0.0],
                maximum=[1.0, 1.0],
                name=f'touch_position_{i}')
      })

  if num_tokens>0:
    action_spec.update({
        'input_token':
          specs.DiscreteArray(
            num_values=num_tokens,
            name="input_token")
      })

    #for i in range(2, num_fingers + 1):
      #action_spec.update({
          #f'input_token_{i}':
            #specs.DiscreteArray(
              #num_values=num_tokens,
              #name="input_token")
        #})

  return action_spec


def base_observation_spec(height: int, width: int) -> Dict[str, specs.Array]:
  """Default observation spec for AndroidEnv.

  Args:
    height: Height of the device screen in pixels.
    width: Width of the device screen in pixels.
  Returns:
    Dict[str, specs.Array]: dict like
      {
        "pixels": Spec for the RGB screenshot of the device. Has shape (H, W, 3)
        "timedelta": Spec for time delta since the last observation (in microseconds).
        "orientation": Spec for the latest orientation in a one-hot representation:
          - [1, 0, 0, 0]: PORTRAIT  (0 degrees)
          - [0, 1, 0, 0]: LANDSCAPE (90 degrees clockwise)
          - [0, 0, 1, 0]: PORTRAIT  (180 degrees) ("upside down")
          - [0, 0, 0, 1]: LANDSCAPE (270 degrees clockwise)
        "view_hierarchy": specs.Array for the view hierarchy observation
      }
  """

  return { 'pixels': specs.Array( shape=(height, width, 3)
                                , dtype=np.uint8
                                , name='pixels'
                                )
         , 'timedelta': specs.Array(shape=(), dtype=np.int64, name='timedelta')
         , 'orientation': specs.Array(shape=np.array([4]), dtype=np.uint8, name='orientation')
         , 'view_hierarchy': specs.Array(shape=(), dtype=np.object_, name="view_hierarchy")
         }


def base_task_extras_spec(task: task_pb2.Task) -> Dict[str, dm_env.specs.Array]:
  """Task extras spec for AndroidEnv, as read from a task_pb2.Task."""

  return {
      spec.name: _convert_spec(spec)
      for spec in task.extras_spec
  }


def _convert_spec(array_spec: task_pb2.ArraySpec) -> specs.Array:
  """Converts ArraySpec proto to dm_env specs.Array."""

  return specs.Array(
      shape=array_spec.shape,
      dtype=_PROTO_DTYPE_TO_NUMPY_DTYPE[array_spec.dtype],
      name=array_spec.name)



# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Prepares and launches the emulator."""

import os
import subprocess
import time
from typing import Optional

from absl import logging
from android_env.components import errors
import grpc

from android_env.proto import emulator_controller_pb2
from android_env.proto import emulator_controller_pb2_grpc
from google.protobuf import empty_pb2

# Period in milliseconds to ping the Emulator gRPC server to keep the connection
# alive. If too frequent, we may get errors such as "Too many pings.", which can
# bring down the process.
_GRPC_KEEPALIVE_MS = 100000

class EmulatorLauncher():
  """Handles launching the emulator."""

  def __init__(
      self,
      local_tmp_dir: str = '/tmp',
      adb_port: Optional[int] = None,
      adb_server_port: Optional[int] = None,
      emulator_console_port: Optional[int] = None,
      grpc_port: int = -1,
      emulator_path: str = '',
      android_sdk_root: str = '',
      avd_name: str = '',
      run_headless: bool = False,
      kvm_device: str = '/dev/kvm',
      gpu_mode: str = 'swiftshader_indirect',
      android_avd_home: str = '',
      startup_wait_time_sec: int = 300,
      writable_system: bool = False,
      proxy_address: Optional[str] = None,
      proxy_port: Optional[str] = None):
    """Installs required files locally and launches the emulator.

    Args:
      local_tmp_dir: Local directory for logs and maybe installing the AVD.
      adb_port: ADB port for the Android device.
      adb_server_port: Port of the ADB server deamon.
      emulator_console_port: Port for telnet communication with the emulator.
      grpc_port: Port for gRPC communication with the emulator.
      emulator_path: Path to the emulator binary.
      android_sdk_root: Root directory of the Android SDK.
      avd_name: Name of the AVD.
      run_headless: Whether to run in headless mode.
      kvm_device: Path to the KVM device.
      gpu_mode: GPU mode override. Supported values are listed at:
        https://developer.android.com/studio/run/emulator-acceleration#accel-graphics
      android_avd_home: Local directory for AVDs.
      startup_wait_time_sec: Timeout for booting the emulator.
      writable_system: Whether to run with a writable /system partition.
      proxy_address: The address for the HTTP proxy.
      proxy_port: The port for the HTTP proxy.
    """
    self._local_tmp_dir: str = local_tmp_dir
    self._adb_port: Optional[int] = adb_port
    self._adb_server_port: Optional[int] = adb_server_port
    self._emulator_console_port: Optional[int] = emulator_console_port
    self._emulator_path: str = emulator_path
    self._android_sdk_root: str = android_sdk_root
    self._avd_name: str = avd_name
    self._run_headless: bool = run_headless
    self._writable_system: bool = writable_system
    self._kvm_device: str = kvm_device
    self._gpu_mode: str = gpu_mode
    self._android_avd_home: str = android_avd_home
    self._startup_wait_time_sec: int = startup_wait_time_sec
    self._grpc_port: int = grpc_port
    self._proxy_address: Optional[str] = proxy_address
    self._proxy_port: Optional[str] = proxy_port

    self._emulator = None
    self._emulator_output = None
    self._emulator_stub = None
    self._is_closed = False

  def launch(self) -> None:
    """Launches the emulator."""

    logging.info('Booting the emulator [%s]', self._emulator_path)

    # Set necessary environment variables.
    base_lib_dir = self._emulator_path[:-8] + 'lib64/'
    ld_library_path = ':'.join([
        base_lib_dir + 'x11/',
        base_lib_dir + 'qt/lib/',
        base_lib_dir + 'gles_swiftshader/',
        base_lib_dir
    ])
    extra_env_vars = {
        'ANDROID_HOME': '',
        'ANDROID_SDK_ROOT': self._android_sdk_root,
        'ANDROID_AVD_HOME': self._android_avd_home,
        'ANDROID_EMULATOR_KVM_DEVICE': self._kvm_device,
        'ANDROID_ADB_SERVER_PORT': str(self._adb_server_port),
        'LD_LIBRARY_PATH': ld_library_path,
        'QT_DEBUG_PLUGINS': '1',
        'QT_XKB_CONFIG_ROOT': str(self._emulator_path[:-8] + 'qt_config/'),
    }
    logging.info('extra_env_vars: %s', str(extra_env_vars))
    env_vars = dict(os.environ).copy()
    env_vars.update(extra_env_vars)

    # Compile command.
    grpc_port = ['-grpc', str(self._grpc_port)] if self._grpc_port >= 0 else []
    run_headless = ['-no-skin', '-no-window'] if self._run_headless else []
    writable_system = ["-writable-system"] if self._writable_system else []
    ports = ['-ports', '%s,%s' % (self._emulator_console_port, self._adb_port)]
    proxy = ["-http-proxy", "http://{:}:{:}".format(self._proxy_address, self._proxy_port)]\
            if self._proxy_address and self._proxy_port else []
    command = [
        self._emulator_path,
        '-no-snapshot',
        '-read-only',
        '-gpu', self._gpu_mode,
        '-no-audio',
        '-verbose',
        '-avd', self._avd_name,
    ] + grpc_port + run_headless + writable_system + ports + proxy
    logging.info('Emulator launch command: %s', ' '.join(command))

    # Prepare logfile.
    emulator_logfile = os.path.join(self._local_tmp_dir, 'emulator_output')
    self._emulator_output = open(emulator_logfile, 'wb')

    # Spawn the emulator process.
    self._emulator = subprocess.Popen(
        command,
        env=env_vars,
        stdout=self._emulator_output,
        stderr=self._emulator_output)

    self._emulator_stub = EmulatorLauncher.create_emulator_stub(self._grpc_port)

    # Wait for the emulator to boot.
    start_time = time.time()
    deadline = start_time + self._startup_wait_time_sec
    success = False
    while time.time() < deadline:
      emu_status = self._emulator_stub.getStatus(empty_pb2.Empty())
      logging.info('Waiting for emulator to start. Emulator uptime: %rms',
                   emu_status.uptime)
      if emu_status.booted:
        success = True
        break
      time.sleep(5.0)

    elapsed_time = time.time() - start_time
    if not success:
      raise errors.SimulatorCrashError(
          'The emulator failed to boot after %r seconds' %
          self._startup_wait_time_sec)

    logging.info('Done booting the emulator (in %f seconds).', elapsed_time)

  def restart(self) -> None:
    logging.info('Restarting the emulator...')
    self._kill_emulator_process()
    self.launch()
    logging.info('Done restarting the emulator.')

  @classmethod
  def create_emulator_stub(
      cls,
      grpc_port: int,
      use_async: bool = False,
  ) -> emulator_controller_pb2_grpc.EmulatorControllerStub:
    """Returns a stub to the EmulatorController service."""
    logging.info('Creating gRPC channel to the emulator on port %r', grpc_port)
    port = f'localhost:{grpc_port}'
    options = [('grpc.max_send_message_length', -1),
               ('grpc.max_receive_message_length', -1),
               ('grpc.keepalive_time_ms', _GRPC_KEEPALIVE_MS)]
    creds = grpc.local_channel_credentials()
    if use_async:
      channel = grpc.aio.secure_channel(port, creds, options=options)
    else:
      channel = grpc.secure_channel(port, creds, options=options)
    grpc.channel_ready_future(channel).result()  # Wait for channel to be ready.
    logging.info('Added gRPC channel for the Emulator on port %s', port)
    return emulator_controller_pb2_grpc.EmulatorControllerStub(channel)

  def get_emulator_stub(
      self) -> emulator_controller_pb2_grpc.EmulatorControllerStub:
    """Returns the EmulatorController stub for the launched emulator."""
    return self._emulator_stub

  def _kill_emulator_process(self) -> None:
    """Shuts down the emulator process."""
    if self._emulator:
      logging.info('Killing the emulator process...')
      self._emulator_stub.setVmState(
          emulator_controller_pb2.VmRunState(
              state=emulator_controller_pb2.VmRunState.RunState.SHUTDOWN))
      logging.info('Will wait 30s for it to finish gracefully...')
      try:
        self._emulator.wait(timeout=30.0)
      except subprocess.TimeoutExpired:
        logging.exception(
            'The emulator process did not finish after 30s. '
            'returncode: %s. Will now try to kill() it.',
            self._emulator.returncode)
        self._emulator.kill()
      self._emulator = None
      self._emulator_output.close()
      logging.info('Done killing the emulator process.')

  def close(self):
    """Clean up launcher files and processes."""
    if not self._is_closed:
      self._kill_emulator_process()
      self._is_closed = True

  def __del__(self):
    self.close()



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wraps the AndroidEnv to expose an OpenAI Gym interface."""

from typing import Any, Dict, Tuple

from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import gym
from gym import spaces
import numpy as np


class GymInterfaceWrapper(base_wrapper.BaseWrapper, gym.Env):
  """AndroidEnv with OpenAI Gym interface."""

  def __init__(self, env: dm_env.Environment):

    base_wrapper.BaseWrapper.__init__(self, env)
    self.spec = None
    self.action_space = self._spec_to_space(self._env.action_spec())
    self.observation_space = self._spec_to_space(self._env.observation_spec())
    self.metadata = {'render.modes': ['rgb_array']}

  def _spec_to_space(self, spec: specs.Array) -> spaces.Space:
    """Converts dm_env specs to OpenAI Gym spaces."""

    if isinstance(spec, list):
      return spaces.Tuple([self._spec_to_space(s) for s in spec])

    if isinstance(spec, dict):
      return spaces.Dict(
          {name: self._spec_to_space(s) for name, s in spec.items()})

    if isinstance(spec, specs.DiscreteArray):
      #return spaces.Box(
          #shape=(),
          #dtype=spec.dtype,
          #low=0,
          #high=spec.num_values-1)
      return spaces.Discrete(spec.num_values)

    if isinstance(spec, specs.BoundedArray):
      return spaces.Box(
          shape=spec.shape,
          dtype=spec.dtype,
          low=spec.minimum.item() if spec.minimum.size==1 else spec.minimum,
          high=spec.maximum.item() if spec.maximum.size==1 else spec.maximum)

    if isinstance(spec, specs.Array):
      return spaces.Box(
          shape=spec.shape,
          dtype=spec.dtype,
          low=-np.inf,
          high=np.inf)

    raise ValueError('Unknown type for specs: {}'.format(spec))

  def render(self, mode='rgb_array'):
    """Renders the environment."""
    if mode == 'rgb_array':
      if self._latest_observation is None:
        return
      return self._latest_observation['pixels']
    else:
      raise ValueError('Only supported render mode is rgb_array.')

  def switch_task(self, index: int) -> np.ndarray:
    timestep = self._env.switch_task(index)
    return timestep.observation

  def reset(self) -> np.ndarray:
    timestep = self._env.reset()
    return timestep.observation

  def step(self, action: Dict[str, int]) -> Tuple[Any, ...]:
    """Take a step in the base environment."""
    timestep = self._env.step(self._process_action(action))
    observation = timestep.observation
    reward = timestep.reward
    done = timestep.step_type == dm_env.StepType.LAST
    info = {'discount': timestep.discount}
    return observation, reward, done, info



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from android_env.proto import emulator_controller_pb2 as android__env_dot_proto_dot_emulator__controller__pb2
from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2


class EmulatorControllerStub(object):
    """An EmulatorController service lets you control the emulator.
    Note that this is currently an experimental feature, and that the
    service definition might change without notice. Use at your own risk!

    We use the following rough conventions:

    streamXXX --> streams values XXX (usually for emulator lifetime). Values
    are updated as soon as they become available.
    getXXX    --> gets a single value XXX
    setXXX    --> sets a single value XXX, does not returning state, these
    usually have an observable lasting side effect.
    sendXXX   --> send a single event XXX, possibly returning state information.
    android usually responds to these events.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.streamSensor = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamSensor',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
                )
        self.getSensor = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getSensor',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
                )
        self.setSensor = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setSensor',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.setPhysicalModel = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setPhysicalModel',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.getPhysicalModel = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getPhysicalModel',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
                )
        self.streamPhysicalModel = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamPhysicalModel',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
                )
        self.setClipboard = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setClipboard',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.ClipData.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.getClipboard = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getClipboard',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.ClipData.FromString,
                )
        self.streamClipboard = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamClipboard',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.ClipData.FromString,
                )
        self.setBattery = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setBattery',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.BatteryState.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.getBattery = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getBattery',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.BatteryState.FromString,
                )
        self.setGps = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setGps',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.GpsState.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.getGps = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getGps',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.GpsState.FromString,
                )
        self.sendFingerprint = channel.unary_unary(
                '/android.emulation.control.EmulatorController/sendFingerprint',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.Fingerprint.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.sendKey = channel.unary_unary(
                '/android.emulation.control.EmulatorController/sendKey',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.KeyboardEvent.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.sendTouch = channel.unary_unary(
                '/android.emulation.control.EmulatorController/sendTouch',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.TouchEvent.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.sendMouse = channel.unary_unary(
                '/android.emulation.control.EmulatorController/sendMouse',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.MouseEvent.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.sendPhone = channel.unary_unary(
                '/android.emulation.control.EmulatorController/sendPhone',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhoneCall.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhoneResponse.FromString,
                )
        self.sendSms = channel.unary_unary(
                '/android.emulation.control.EmulatorController/sendSms',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.SmsMessage.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhoneResponse.FromString,
                )
        self.getStatus = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getStatus',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.EmulatorStatus.FromString,
                )
        self.getScreenshot = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getScreenshot',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.ImageFormat.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.Image.FromString,
                )
        self.streamScreenshot = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamScreenshot',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.ImageFormat.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.Image.FromString,
                )
        self.streamAudio = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamAudio',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.AudioFormat.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.AudioPacket.FromString,
                )
        self.injectAudio = channel.stream_unary(
                '/android.emulation.control.EmulatorController/injectAudio',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.AudioPacket.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.getLogcat = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getLogcat',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.FromString,
                )
        self.streamLogcat = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamLogcat',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.FromString,
                )
        self.setVmState = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setVmState',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.VmRunState.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.getVmState = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getVmState',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.VmRunState.FromString,
                )
        self.setDisplayConfigurations = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setDisplayConfigurations',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.FromString,
                )
        self.getDisplayConfigurations = channel.unary_unary(
                '/android.emulation.control.EmulatorController/getDisplayConfigurations',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.FromString,
                )
        self.streamNotification = channel.unary_stream(
                '/android.emulation.control.EmulatorController/streamNotification',
                request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
                response_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.Notification.FromString,
                )
        self.rotateVirtualSceneCamera = channel.unary_unary(
                '/android.emulation.control.EmulatorController/rotateVirtualSceneCamera',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.RotationRadian.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.setVirtualSceneCameraVelocity = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setVirtualSceneCameraVelocity',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.Velocity.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )
        self.setPosture = channel.unary_unary(
                '/android.emulation.control.EmulatorController/setPosture',
                request_serializer=android__env_dot_proto_dot_emulator__controller__pb2.Posture.SerializeToString,
                response_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                )


class EmulatorControllerServicer(object):
    """An EmulatorController service lets you control the emulator.
    Note that this is currently an experimental feature, and that the
    service definition might change without notice. Use at your own risk!

    We use the following rough conventions:

    streamXXX --> streams values XXX (usually for emulator lifetime). Values
    are updated as soon as they become available.
    getXXX    --> gets a single value XXX
    setXXX    --> sets a single value XXX, does not returning state, these
    usually have an observable lasting side effect.
    sendXXX   --> send a single event XXX, possibly returning state information.
    android usually responds to these events.
    """

    def streamSensor(self, request, context):
        """Set the sensor data
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getSensor(self, request, context):
        """Get the sensor data
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setSensor(self, request, context):
        """Stream the sensor data
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setPhysicalModel(self, request, context):
        """Set the physical model, this is likely the one you are
        looking for when you wish to modify the device state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getPhysicalModel(self, request, context):
        """Get the physical model
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def streamPhysicalModel(self, request, context):
        """Stream the physical model
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setClipboard(self, request, context):
        """Atomically set the current primary clipboard data.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getClipboard(self, request, context):
        """Atomically get the current primary clipboard data.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def streamClipboard(self, request, context):
        """Streams the current data on the clipboard. This will immediately produce
        a result with the current state of the clipboard after which the stream
        will block and wait until a new clip event is available from the guest.
        Calling the setClipboard method above will not result in generating a clip
        event. It is possible to lose clipboard events if the clipboard updates
        very rapidly.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setBattery(self, request, context):
        """Set the battery to the given state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getBattery(self, request, context):
        """Get the battery to the given state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setGps(self, request, context):
        """Set the state of the gps, gps support will only work
        properly if:

        - no location ui is active. That is the emulator
        is launched in headless mode (-no-window) or the location
        ui is disabled (-no-location-ui).
        - the passiveUpdate is set to false. Setting this to false
        will disable/break the LocationUI.

        Keep in mind that android usually only samples the gps at 1 hz.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getGps(self, request, context):
        """Gets the latest gps state as delivered by the setGps call, or location ui
        if active.

        Note: this is not necessarily the actual gps coordinate visible at the
        time, due to gps sample frequency (usually 1hz).
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendFingerprint(self, request, context):
        """Simulate a touch event on the finger print sensor.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendKey(self, request, context):
        """Send a keyboard event. Translating the event.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendTouch(self, request, context):
        """Send touch events. Note that mouse events can be simulated by touch events.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendMouse(self, request, context):
        """Send mouse events.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendPhone(self, request, context):
        """Make a phone call.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def sendSms(self, request, context):
        """Sends an sms message to the emulator.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getStatus(self, request, context):
        """Retrieve the status of the emulator. This will contain general
        hardware information, and whether the device has booted or not.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getScreenshot(self, request, context):
        """Gets an individual screenshot in the desired format.

        The image will be scaled to the desired ImageFormat, while maintaining
        the aspect ratio. The returned image will never exceed the provided width
        and height. Not setting the width or height (i.e. they are 0) will result
        in using the device width and height.

        The resulting image will be properly oriented and can be displayed
        directly without post processing. For example, if the device has a
        1080x1920 screen and is in landscape mode and called with no width or
        height parameter, it will return an 1920x1080 image.

        This method will return an empty image if the display is not visible.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def streamScreenshot(self, request, context):
        """Streams a series of screenshots in the desired format.
        A new frame will be delivered whenever the device produces a new frame.
        (Beware that this can produce a significant amount of data, and that
        certain translations are (png transform) can be costly).

        If the requested display is not visible it will send a single empty image
        and wait start producing images once the display becomes active, again
        producing a single empty image when the display becomes inactive.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def streamAudio(self, request, context):
        """Streams a series of audio packets in the desired format.
        A new frame will be delivered whenever the emulated device
        produces a new audio frame. You can expect packets to be
        delivered in intervals of 20-30ms.

        Be aware that this can block when the emulator does not
        produce any audio whatsoever!
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def injectAudio(self, request_iterator, context):
        """Injects a series of audio packets to the android microphone.
        A new frame will be delivered whenever the emulated device
        requests a new audio frame. Audio is usually delivered at a rate
        that the emulator is requesting frames. Audio will be stored in a
        temporary buffer that can hold 500ms of audio.

        Note: Currently the emulator will downsample to 16khz.

        -  INVALID_ARGUMENT (code 3) The sampling rate was too high
        -  INVALID_ARGUMENT (code 3) The audio packet was too large to handle.
        -  FAILED_PRECONDITION (code 9) If there was a microphone registered already.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getLogcat(self, request, context):
        """Returns the last 128Kb of logcat output from the emulator
        Note that parsed logcat messages are only available after L (Api >23).
        it is possible that the logcat buffer gets overwritten, or falls behind.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def streamLogcat(self, request, context):
        """Streams the logcat output from the emulator. The first call
        can retrieve up to 128Kb. This call will not return.
        Note that parsed logcat messages are only available after L (Api >23)
        it is possible that the logcat buffer gets overwritten, or falls behind.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setVmState(self, request, context):
        """Transition the virtual machine to the desired state. Note that
        some states are only observable. For example you cannot transition
        to the error state.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getVmState(self, request, context):
        """Gets the state of the virtual machine.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setDisplayConfigurations(self, request, context):
        """Atomically changes the current multi-display configuration.
        After this call the given display configurations will be activated. You
        can only update secondary displays. Displays with id 0 will be ignored.

        This call can result in the removal or addition of secondary displays, the final
        display state can be observed by the returned configuration.

        The following gRPC error codes can be returned:
        -  FAILED_PRECONDITION (code 9) if the AVD does not support a configurable
        secondary display.
        -  INVALID_ARGUMENT (code 3) if:
        - The same display id is defined multiple times.
        - The display configurations are outside valid ranges
        (see DisplayConfiguration)
        -  INTERNAL (code 13) if there was an internal emulator failure.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def getDisplayConfigurations(self, request, context):
        """Returns all currently valid logical displays.
        The gRPC error code FAILED_PRECONDITION (code 9) is returned if the AVD
        does not support a configurable secondary display.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def streamNotification(self, request, context):
        """Notifies client of the following changes:

        - Virtual scene camera status change.
        - Display configuration changes from extended ui. This will only be fired
        if the user makes modifications the extended displays through the extended
        control tab.

        Note that this method will send the initial virtual scene state immediately.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def rotateVirtualSceneCamera(self, request, context):
        """RotationRadian is relative to the camera's current orientation.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setVirtualSceneCameraVelocity(self, request, context):
        """Velocity is absolute
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def setPosture(self, request, context):
        """Set foldable posture
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_EmulatorControllerServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'streamSensor': grpc.unary_stream_rpc_method_handler(
                    servicer.streamSensor,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
            ),
            'getSensor': grpc.unary_unary_rpc_method_handler(
                    servicer.getSensor,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
            ),
            'setSensor': grpc.unary_unary_rpc_method_handler(
                    servicer.setSensor,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'setPhysicalModel': grpc.unary_unary_rpc_method_handler(
                    servicer.setPhysicalModel,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'getPhysicalModel': grpc.unary_unary_rpc_method_handler(
                    servicer.getPhysicalModel,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
            ),
            'streamPhysicalModel': grpc.unary_stream_rpc_method_handler(
                    servicer.streamPhysicalModel,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
            ),
            'setClipboard': grpc.unary_unary_rpc_method_handler(
                    servicer.setClipboard,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.ClipData.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'getClipboard': grpc.unary_unary_rpc_method_handler(
                    servicer.getClipboard,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.ClipData.SerializeToString,
            ),
            'streamClipboard': grpc.unary_stream_rpc_method_handler(
                    servicer.streamClipboard,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.ClipData.SerializeToString,
            ),
            'setBattery': grpc.unary_unary_rpc_method_handler(
                    servicer.setBattery,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.BatteryState.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'getBattery': grpc.unary_unary_rpc_method_handler(
                    servicer.getBattery,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.BatteryState.SerializeToString,
            ),
            'setGps': grpc.unary_unary_rpc_method_handler(
                    servicer.setGps,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.GpsState.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'getGps': grpc.unary_unary_rpc_method_handler(
                    servicer.getGps,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.GpsState.SerializeToString,
            ),
            'sendFingerprint': grpc.unary_unary_rpc_method_handler(
                    servicer.sendFingerprint,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.Fingerprint.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'sendKey': grpc.unary_unary_rpc_method_handler(
                    servicer.sendKey,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.KeyboardEvent.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'sendTouch': grpc.unary_unary_rpc_method_handler(
                    servicer.sendTouch,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.TouchEvent.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'sendMouse': grpc.unary_unary_rpc_method_handler(
                    servicer.sendMouse,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.MouseEvent.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'sendPhone': grpc.unary_unary_rpc_method_handler(
                    servicer.sendPhone,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.PhoneCall.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhoneResponse.SerializeToString,
            ),
            'sendSms': grpc.unary_unary_rpc_method_handler(
                    servicer.sendSms,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.SmsMessage.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.PhoneResponse.SerializeToString,
            ),
            'getStatus': grpc.unary_unary_rpc_method_handler(
                    servicer.getStatus,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.EmulatorStatus.SerializeToString,
            ),
            'getScreenshot': grpc.unary_unary_rpc_method_handler(
                    servicer.getScreenshot,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.ImageFormat.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.Image.SerializeToString,
            ),
            'streamScreenshot': grpc.unary_stream_rpc_method_handler(
                    servicer.streamScreenshot,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.ImageFormat.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.Image.SerializeToString,
            ),
            'streamAudio': grpc.unary_stream_rpc_method_handler(
                    servicer.streamAudio,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.AudioFormat.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.AudioPacket.SerializeToString,
            ),
            'injectAudio': grpc.stream_unary_rpc_method_handler(
                    servicer.injectAudio,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.AudioPacket.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'getLogcat': grpc.unary_unary_rpc_method_handler(
                    servicer.getLogcat,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.SerializeToString,
            ),
            'streamLogcat': grpc.unary_stream_rpc_method_handler(
                    servicer.streamLogcat,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.SerializeToString,
            ),
            'setVmState': grpc.unary_unary_rpc_method_handler(
                    servicer.setVmState,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.VmRunState.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'getVmState': grpc.unary_unary_rpc_method_handler(
                    servicer.getVmState,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.VmRunState.SerializeToString,
            ),
            'setDisplayConfigurations': grpc.unary_unary_rpc_method_handler(
                    servicer.setDisplayConfigurations,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.SerializeToString,
            ),
            'getDisplayConfigurations': grpc.unary_unary_rpc_method_handler(
                    servicer.getDisplayConfigurations,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.SerializeToString,
            ),
            'streamNotification': grpc.unary_stream_rpc_method_handler(
                    servicer.streamNotification,
                    request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
                    response_serializer=android__env_dot_proto_dot_emulator__controller__pb2.Notification.SerializeToString,
            ),
            'rotateVirtualSceneCamera': grpc.unary_unary_rpc_method_handler(
                    servicer.rotateVirtualSceneCamera,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.RotationRadian.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'setVirtualSceneCameraVelocity': grpc.unary_unary_rpc_method_handler(
                    servicer.setVirtualSceneCameraVelocity,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.Velocity.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
            'setPosture': grpc.unary_unary_rpc_method_handler(
                    servicer.setPosture,
                    request_deserializer=android__env_dot_proto_dot_emulator__controller__pb2.Posture.FromString,
                    response_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'android.emulation.control.EmulatorController', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class EmulatorController(object):
    """An EmulatorController service lets you control the emulator.
    Note that this is currently an experimental feature, and that the
    service definition might change without notice. Use at your own risk!

    We use the following rough conventions:

    streamXXX --> streams values XXX (usually for emulator lifetime). Values
    are updated as soon as they become available.
    getXXX    --> gets a single value XXX
    setXXX    --> sets a single value XXX, does not returning state, these
    usually have an observable lasting side effect.
    sendXXX   --> send a single event XXX, possibly returning state information.
    android usually responds to these events.
    """

    @staticmethod
    def streamSensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamSensor',
            android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getSensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getSensor',
            android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setSensor(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setSensor',
            android__env_dot_proto_dot_emulator__controller__pb2.SensorValue.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setPhysicalModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setPhysicalModel',
            android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getPhysicalModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getPhysicalModel',
            android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def streamPhysicalModel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamPhysicalModel',
            android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.PhysicalModelValue.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setClipboard(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setClipboard',
            android__env_dot_proto_dot_emulator__controller__pb2.ClipData.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getClipboard(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getClipboard',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.ClipData.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def streamClipboard(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamClipboard',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.ClipData.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setBattery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setBattery',
            android__env_dot_proto_dot_emulator__controller__pb2.BatteryState.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getBattery(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getBattery',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.BatteryState.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setGps(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setGps',
            android__env_dot_proto_dot_emulator__controller__pb2.GpsState.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getGps(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getGps',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.GpsState.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def sendFingerprint(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/sendFingerprint',
            android__env_dot_proto_dot_emulator__controller__pb2.Fingerprint.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def sendKey(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/sendKey',
            android__env_dot_proto_dot_emulator__controller__pb2.KeyboardEvent.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def sendTouch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/sendTouch',
            android__env_dot_proto_dot_emulator__controller__pb2.TouchEvent.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def sendMouse(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/sendMouse',
            android__env_dot_proto_dot_emulator__controller__pb2.MouseEvent.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def sendPhone(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/sendPhone',
            android__env_dot_proto_dot_emulator__controller__pb2.PhoneCall.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.PhoneResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def sendSms(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/sendSms',
            android__env_dot_proto_dot_emulator__controller__pb2.SmsMessage.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.PhoneResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getStatus(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getStatus',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.EmulatorStatus.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getScreenshot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getScreenshot',
            android__env_dot_proto_dot_emulator__controller__pb2.ImageFormat.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.Image.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def streamScreenshot(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamScreenshot',
            android__env_dot_proto_dot_emulator__controller__pb2.ImageFormat.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.Image.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def streamAudio(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamAudio',
            android__env_dot_proto_dot_emulator__controller__pb2.AudioFormat.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.AudioPacket.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def injectAudio(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(request_iterator, target, '/android.emulation.control.EmulatorController/injectAudio',
            android__env_dot_proto_dot_emulator__controller__pb2.AudioPacket.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getLogcat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getLogcat',
            android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def streamLogcat(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamLogcat',
            android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.LogMessage.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setVmState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setVmState',
            android__env_dot_proto_dot_emulator__controller__pb2.VmRunState.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getVmState(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getVmState',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.VmRunState.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setDisplayConfigurations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setDisplayConfigurations',
            android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def getDisplayConfigurations(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/getDisplayConfigurations',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.DisplayConfigurations.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def streamNotification(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/android.emulation.control.EmulatorController/streamNotification',
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            android__env_dot_proto_dot_emulator__controller__pb2.Notification.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def rotateVirtualSceneCamera(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/rotateVirtualSceneCamera',
            android__env_dot_proto_dot_emulator__controller__pb2.RotationRadian.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setVirtualSceneCameraVelocity(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setVirtualSceneCameraVelocity',
            android__env_dot_proto_dot_emulator__controller__pb2.Velocity.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def setPosture(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/android.emulation.control.EmulatorController/setPosture',
            android__env_dot_proto_dot_emulator__controller__pb2.Posture.SerializeToString,
            google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# vim: set tabstop=2 shiftwidth=2:
# coding=utf-8
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Coordinator handles interaction between internal components of AndroidEnv."""

import copy
import socket
import time
from typing import Any, Optional, Iterable, Union
from typing import Dict, Tuple, List

from absl import logging
from android_env.components import action_type as action_type_lib
from android_env.components import errors
from android_env.components import specs
from android_env.components import task_manager as task_manager_lib
from android_env.components.simulators import base_simulator
import dm_env
import numpy as np
import lxml.etree


class Coordinator():
  """Handles interaction between internal components of AndroidEnv."""

  def __init__( self
              , simulator: base_simulator.BaseSimulator
              , task_managers: Iterable[task_manager_lib.TaskManager]
              , step_timeout_sec: int = 10
              , max_steps_per_sec: float = 15.0
              , periodic_restart_time_min: float = 0.0
              , force_simulator_launch: bool = True
              , with_view_hierarchy: bool = False
  ):
    #  method `__init__` {{{ # 
    """Handles communication between AndroidEnv and its components.

    Args:
      simulator: A BaseSimulator instance.
      task_managers: iterable of task_manager_lib.TaskManager as the
        TaskManagers, responsible for coordinating RL tasks.
      step_timeout_sec: Timeout in seconds between steps. If step is not called
        within that time, the episode will reset at the next step. Set to 0 to
        disable.
      max_steps_per_sec: Maximum steps per second. If the simulator is
        faster, the Coordinator will wait before returning an observation.
        If <=0.0, this functionality is turned off.
      periodic_restart_time_min: Time between periodic restarts in minutes.
        If > 0.0, will trigger a simulator restart at the end of the next
        episode once the time has been reached.
      force_simulator_launch: Forces the simulator to relaunch even if it is
        already launched.
      with_view_hierarchy (bool): if the view hierarchy should be included in
        the observation
    """

    self._simulator: base_simulator.BaseSimulator = simulator
    self._task_manager_list: List[task_manager_lib.TaskManager] = list(task_managers)
    self._task_manager_index: int = 0
    self._task_manager: task_manager_lib.TaskManager = self._task_manager_list[self._task_manager_index]

    self._step_timeout_sec = step_timeout_sec
    self._max_steps_per_sec = max_steps_per_sec
    self._periodic_restart_time_min = periodic_restart_time_min
    self._force_simulator_launch = force_simulator_launch
    self._with_view_hierarchy: bool = with_view_hierarchy

    # Logging settings.
    self._log_dict = {
        'total_steps': 0,
        'episode_steps': 0,
        'restart_count': 0,
        'restart_count_periodic': 0,
        'restart_count_setup_steps': 0,
        'restart_count_reset_steps': 0,
        'restart_count_simulator_launch': 0,
        'restart_count_simulator_reset': 0,
        'restart_count_execute_action': 0,
        'restart_count_fetch_observation': 0,
        'restart_count_step_timeout': 0,
    }

    # Initialize counters.
    self._should_restart = False
    self._latest_observation_time = None
    self._simulator_start_time = None

    logging.info('Starting the simulator...')
    self._restart_simulator()
    #  }}} method `__init__` # 

  #  Informational Interfaces {{{ # 
  def action_spec(self) -> Dict[str, dm_env.specs.Array]:
    num_tokens = self._task_manager.nb_tokens()
    num_fingers = self._simulator.num_fingers()
    return specs.base_action_spec(num_tokens=num_tokens, num_fingers=num_fingers)

  def observation_spec(self) -> Dict[str, dm_env.specs.Array]:
    screen_dims = self._simulator.screen_dimensions()
    return specs.base_observation_spec(
        height=screen_dims[0], width=screen_dims[1])

  def task_extras_spec(self) -> Dict[str, dm_env.specs.Array]:
    return specs.base_task_extras_spec(task=self._task_manager.task())

  def command(self) -> List[str]:
    """
    return list of str
    """

    return self._task_manager.command()
  def vocabulary(self) -> List[str]:
    """
    return list of str
    """

    return self._task_manager.vocabulary()

  @property
  def nb_tasks(self) -> int:
    return len(self._task_manager_list)
  @property
  def task_index(self) -> int:
    return self._task_manager_index
  @property
  def task_id(self) -> str:
    return self._task_manager.task().id
  @property
  def task_name(self) -> str:
    return self._task_manager.task().name
  @property
  def task_description(self) -> str:
    return self._task_manager.task().description
  #  }}} Informational Interfaces # 

  #  Reset Interfaces {{{ # 
  def reset_environment_state(self):
    """Resets the state of the simulation for a new RL episode.

    This involves resetting relevant counters and performing reset steps
    specific to the running task (e.g. restarting an application). A lift
    action is also performed to ensure that sequential touches are not
    interconnected between separate RL episodes.
    """

    # Restart the simulation if neccessary.
    if self._should_restart or self._should_periodic_restart():
      self._restart_simulator()

    # Reset counters.
    self._latest_observation_time = None
    for key in self._log_dict:
      if key.startswith('episode'):
        self._log_dict[key] = 0.0

    # Execute a lift action before resetting the task.
    self._lift_all_fingers()

    # Reset the task.
    try:
      self._task_manager.reset_task()
      self._simulator.update_device_orientation() # ZDY_COMMENT: device orientation updated only during reset
    except errors.StepCommandError:
      logging.exception('Failed to reset the task. Restarting simulator.')
      self._log_dict['restart_count_simulator_reset'] += 1
      self._should_restart = True

  def _lift_all_fingers(self) -> None:
    """Performs a lift action with every finger."""
    lift_action = {
        'action_type': np.array(action_type_lib.ActionType.LIFT),
        'touch_position': np.array([0, 0]),
    }
    for i in range(2, self._simulator.num_fingers()+1):
      lift_action.update({
          f'action_type_{i}': np.array(action_type_lib.ActionType.LIFT),
          f'touch_position_{i}': np.array([0, 0]),
      })
    self._send_action_to_simulator(lift_action)

  def _should_periodic_restart(self) -> bool:
    """Checks if it is time to restart the simulator.

    If a periodic restart time was specified, the Coordinator will re-launch
    the simulator at regular time intervals. This helps to make sure that the
    simulator is not is a stale state even if the environment has been running
    for a significant amount of time.

    Returns:
      Boolean indicating if it is time to restart the simulator.
    """

    if self._periodic_restart_time_min and self._simulator_start_time:
      sim_alive_time = (time.time() - self._simulator_start_time) / 60.0
      logging.info('Simulator has been running for %f mins', sim_alive_time)
      if sim_alive_time > self._periodic_restart_time_min:
        logging.info('Maximum alive time reached. Restarting simulator.')
        self._log_dict['restart_count_periodic'] += 1
        return True
    return False

  def _restart_simulator(self, max_retries: int = 3):
    """Restarts the simulation.

    Closes and re-launches the system, restarting the simulator process and
    reinitializing the task in the newly started simulator.

    Args:
      max_retries: Number of times to attempt a restart before raising an error.
    """

    # Reset counters.
    self._should_restart = False

    # Attempt to restart the system a given number of times.
    num_tries = 1
    while True:
      if num_tries > max_retries:
        logging.error('Maximum number of restarts reached.')
        raise errors.TooManyRestartsError
      logging.info('Simulator launch attempt %d of %d', num_tries, max_retries)

      # Launch the simulator (will restart if already launched).
      try:
        if self._force_simulator_launch or not self._simulator.is_launched():
          self._task_manager.pause_task()
          self._simulator.launch()
          self._simulator_start_time = time.time()
        if not hasattr(self, "_adb_controller") or self._adb_controller is None:
          self._adb_controller = self._simulator.create_adb_controller()
        log_stream = self._simulator.get_log_stream()
      except errors.AdbControllerError:
        logging.error('Error launching the simulator.')
        self._log_dict['restart_count_simulator_launch'] += 1
        num_tries += 1
        continue

      # Start the task.
      for t_mng in self._task_manager_list:
        t_mng.clear_setup_flag()
      try:
        self._task_manager.setup_task(
            adb_controller=self._adb_controller,
            #emulator_stub=self._simulator.get_emulator_stub(), # zdy
            #image_format=self._simulator.image_format, # zdy
            log_stream=log_stream)
      except errors.StepCommandError:
        logging.error('Failed to set up the task. Restarting simulator.')
        self._log_dict['restart_count_setup_steps'] += 1
        num_tries += 1
        continue

      # Restart was successful.
      break

  def add_task_manager(self, task_manager: task_manager_lib.TaskManager):
    #  method `add_task_manager` {{{ # 
    """
    Args:
        task_manager: task_manager_lib.TaskManager
    """

    self._task_manager_list.append(task_manager)
    #  }}} method `add_task_manager` # 

  def switch_task_manager(self, index: int):
    #  method `change_task_manager` {{{ # 
    self._task_manager_index = index
    self._task_manager = self._task_manager_list[self._task_manager_index]

    if not self._task_manager.setup_flag():
      try:
        self._task_manager.setup_task(
            adb_controller=self._adb_controller,
            #emulator_stub=self._simulator.get_emulator_stub(), # zdy
            #image_format=self._simulator.image_format, # zdy
            log_stream=self._simulator.get_log_stream())
      except errors.StepCommandError:
        logging.error('Failed to set up the task. Restarting simulator.')
        self._log_dict['restart_count_setup_steps'] += 1
        self._should_restart = True

    self.reset_environment_state()
    #  }}} method `change_task_manager` # 
  #  }}} Reset Interfaces # 

  def execute_action( self
                    , action: Optional[Dict[str, np.ndarray]]
                    ) -> Tuple[ Optional[ Dict[ str
                                              , Union[ np.ndarray
                                                     , Optional[lxml.etree.Element]
                                                     ]
                                              ]
                                        ]
                              , float
                              , Dict[str, Any]
                              , List[str]
                              , bool
                              ]:
    """Executes the selected action and returns transition info.

    Args:
      action: Selected action to perform on the simulated Android device.
    Returns:
      observation: Pixel observations as displayed on the screen.
      reward: Total reward collected since the last call.
      extras: Task extras observed since the last call.
      instructions: Task instructions received since the last call.
      episode_end: Boolean indicating if the RL episode should be terminated.
    """

    # Increment counters.
    self._task_manager.increment_steps()
    if action is not None:
      self._log_dict['total_steps'] += 1
      self._log_dict['episode_steps'] += 1

    # If a restart is neccessary, end the episode.
    if self._should_restart or self._check_timeout():
      return None, 0.0, {}, [], True

    if action is not None:

      if "response" in action\
          and action["response"] is not None\
          and action["response"] != "":
        self._task_manager.receive_response(action["response"])
        pass

      if action["action_type"].item() == action_type_lib.ActionType.TEXT:
        self._send_action_to_taskmanager(action)
      elif action['action_type'].item() != action_type_lib.ActionType.REPEAT:
        self._send_action_to_simulator(action)
      vh: bool = action["action_type"].item()==action_type_lib.ActionType.LIFT
    else:
      vh: bool = True
    vh = vh and self._with_view_hierarchy

    # Sleep to maintain a steady interaction rate.
    if self._max_steps_per_sec > 0.0:
      self._wait_for_next_frame()
    #time.sleep(1)

    # Read necessary transition information and return it to the agent.
    try:
      self._latest_observation_time = time.time()
      observation = self._simulator.get_observation()

      self._task_manager.snapshot_events(observation["pixels"])
      reward, view_hierarchy = self._task_manager.get_current_reward(vh) # zdy
      if self._with_view_hierarchy:
        observation["view_hierarchy"] = view_hierarchy
      task_extras = self._task_manager.get_current_extras()
      instructions = self._task_manager.get_current_instructions()
      episode_end = self._task_manager.check_if_episode_ended(self._with_view_hierarchy)
      self._task_manager.clear_events()
      return observation, reward, task_extras, instructions, episode_end
    except (errors.ReadObservationError, socket.error):
      logging.exception('Unable to fetch observation. Restarting simulator.')
      self._log_dict['restart_count_fetch_observation'] += 1
      self._should_restart = True
      return None, 0.0, {}, [], True

  def _send_action_to_taskmanager(self, action: Dict[str, np.ndarray]):
    #  method `_send_action_to_taskmanager` {{{ # 
    """
    action - dict like
      {
        "action_type": scalar array of int32 with value 3,
        "input_token": scalar array of int32
      }
    """

    self._task_manager.send_token(action["input_token"].item())
    #  }}} method `_send_action_to_taskmanager` # 

  def _send_action_to_simulator(self, action: Dict[str, np.ndarray]) -> None:
    """Sends the selected action to the simulator.

    The simulator will interpret the action as a touchscreen event and perform
    it accordingly. The effect this action triggers in the Android OS will be
    determined by the currently running application.

    Args:
      action: action which will get interpreted as a touchscreen event.
    """

    try:
      self._simulator.send_action(action)
    except (socket.error, errors.SendActionError):
      logging.exception('Unable to execute action. Restarting simulator.')
      self._log_dict['restart_count_execute_action'] += 1
      self._should_restart = True

  def _check_timeout(self) -> bool:
    """Checks if timeout between steps have exceeded.

    If too much time has passed since the last step was performed, it is assumed
    that the simulation is in a bad state, so the Coordinator will re-launch
    the simulator to make sure interaction proceeds from a clean state.

    Returns:
      Boolean indicating if the step timeout limit has been reached.
    """

    if not self._with_view_hierarchy\
        and self._step_timeout_sec and self._latest_observation_time:
      time_since_last_obs = self._get_time_since_last_observation()
      if time_since_last_obs > self._step_timeout_sec:
        logging.exception('Time between steps exceeded %f',
                          self._step_timeout_sec)
        self._log_dict['restart_count_step_timeout'] += 1
        self._should_restart = True
        return True
    return False

  def _wait_for_next_frame(self) -> None:
    """Pauses the environment so that the interaction is around 1/FPS."""

    time_since_observation = self._get_time_since_last_observation()
    time_to_wait = 1. / self._max_steps_per_sec - time_since_observation
    if time_to_wait > 0.0:
      time.sleep(time_to_wait)

  def _get_time_since_last_observation(self) -> float:
    """Computes time passed since the last observation was fetched."""

    if self._latest_observation_time is not None:
      return time.time() - self._latest_observation_time
    else:
      return np.inf

  def get_logs(self) -> Dict[str, Any]:
    """Returns internal counter values."""

    log_dict = copy.deepcopy(self._log_dict)
    log_dict.update(self._task_manager.log_dict())
    return log_dict

  def close(self):
    """Cleans up the state of this Coordinator."""

    if hasattr(self, '_task_manager'):
      self._task_manager.close()
    if hasattr(self, '_simulator'):
      self._simulator.close()



# coding=utf-8
# vim: set tabstop=2 shiftwidth=2:
# Copyright 2023 SJTU X-Lance Lab
# 
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
# 
#     http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# 
# Revised by Danyang Zhang @X-Lance based on
# 
# Copyright 2021 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Converts pixel observation to from int to float32 between 0.0 and 1.0."""

from typing import Dict

from android_env.components import utils
from android_env.wrappers import base_wrapper
import dm_env
from dm_env import specs
import numpy as np

imgnet_mean = np.array([0.485, 0.456, 0.406])
imgnet_std = np.array([0.229, 0.224, 0.225])
def normalize_pixels(pixels: np.ndarray, imgnet_norm: bool = False) -> np.ndarray:
  #  function `normalize_pixel` {{{ # 
  """
  Args:
      pixels: array of uint8 with shape (height, width, nb_channels)
      imgnet_norm: bool

  Returns:
      array of float32 with the same shape as `pixels`
  """

  float_pixels = np.clip(pixels, 0, 255).astype(np.float32)/255.
  if imgnet_norm:
    float_pixels -= imgnet_mean
    float_pixels /= imgnet_std
  return float_pixels
    #  }}} function `normalize_pixel` # 

class FloatPixelsWrapper(base_wrapper.BaseWrapper):
  """Wraps AndroidEnv for Panultimate agent."""

  def __init__(self, env: dm_env.Environment, imgnet_norm: bool = False):
    super().__init__(env)
    self._should_convert_int_to_float = np.issubdtype(
        self._env.observation_spec()['pixels'].dtype, np.integer)
    self._imgnet_norm = imgnet_norm

  def _process_observation(
      self, observation: Dict[str, np.ndarray]
  ) -> Dict[str, np.ndarray]:
    if self._should_convert_int_to_float:
      #float_pixels = utils.convert_int_to_float(
          #observation['pixels'],
          #self._env.observation_spec()['pixels'],
          #np.float32)
      float_pixels = normalize_pixels(observation["pixels"], self._imgnet_norm)
      observation['pixels'] = float_pixels
    return observation

  def _process_timestep(self, timestep: dm_env.TimeStep) -> dm_env.TimeStep:
    step_type, reward, discount, observation = timestep
    return dm_env.TimeStep(
        step_type=step_type,
        reward=reward,
        discount=discount,
        observation=self._process_observation(observation))

  def reset(self) -> dm_env.TimeStep:
    timestep = self._env.reset()
    return self._process_timestep(timestep)

  def step(self, action: Dict[str, np.ndarray]) -> dm_env.TimeStep:
    timestep = self._env.step(action)
    return self._process_timestep(timestep)

  def observation_spec(self) -> Dict[str, specs.Array]:
    if self._should_convert_int_to_float:
      observation_spec = self._env.observation_spec()
      observation_spec['pixels'] = specs.BoundedArray(
          shape=self._env.observation_spec()['pixels'].shape,
          dtype=np.float32,
          minimum=0.0,
          maximum=1.0,
          name=self._env.observation_spec()['pixels'].name)
      return observation_spec
    return self._env.observation_spec()
